#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import math
import numpy as np
import scipy.spatial

import collections
import heapq

import Environment as env
import SamplingPlanner as sp

from matplotlib import pyplot as plt

class PRMPlanner(sp.SamplingPlanner):

    def create_sub_path(self, start, end):

        t1 = start[:3,3]
        t2 = end[:3,3]

        R1 = start[:3,:3]
        R2 = end[:3,:3]

        t_len = sp.norm_translation(t1,t2)
        t_step = self.collision_resolution_translation
        step_no_t = math.ceil ( t_len / t_step ) 

        R_len = sp.norm_rotation_R(R1,R2)
        R_step = self.collision_resolution_rotation
        step_no_R = math.ceil ( R_len / R_step ) 

        step_no_max = max(step_no_R, step_no_t)

        pth = sp.slerp_se3(start, end, step_no_max)

        return pth   

    def check_collision(self, pth, environment, verbose = False):
        ret = False
        for se3 in pth:
            ret = environment.check_robot_collision(se3)
            if ret:
                if verbose:
                    print ('collision at %s' % str(se3))
                break
        return ret    

    def is_valid_edge(self, start, end, environment, verbose = False):
        """
        Is and oriented edge between two configurations valid? 

        Parameters
        ----------
        start: numpy array (4x4)
            start configuration of the robot in SE(3) coordinates
        end: numpy array (4x4)
            end configuration of the robot in SE(3) coordinates
        environment: Environment
            Map of the environment that provides collision checking 
        verbose: boolean
             if True then verbose

        Returns
        -------
        boolean
            True if the edge is valid
        """


        ret = True

        t1 = start[:3,3]
        t2 = end[:3,3]
        R1 = start[:3,:3]
        R2 = end[:3,:3]

        if verbose:
            print ('t1 : %s' % str(t1))
            print ('t2 : %s' % str(t2))
            print ('translation len : %.6f' % sp.norm_translation(t1,t2))
            print ('rotation    len : %.6f' % sp.norm_rotation_R(R1,R2))

        if sp.norm_translation(t1,t2) > self.max_norm_translation:
            ret = False
            if verbose:
                print ( 'translation fail')
        elif sp.norm_rotation_R(R1,R2) > self.max_norm_rotation:    
            ret = False
            if verbose:
                print ( 'rotation fail')
        elif self.check_collision(self.create_sub_path(start,end),environment):
            ret = False
            if verbose:
                print ( 'collision fail')

        return ret

    def plan(self, environment, start, goal, n_samples = 300):
        """
        Method to plan the path

        Parameters
        ----------
        environment: Environment
            Map of the environment that provides collision checking 
        start: numpy array (4x4)
            start configuration of the robot in SE(3) coordinates
        goal: numpy array (4x4)
            goal configuration of the robot in SE(3) coordinates
        n_samples: int
            number of C_free samples

        Returns
        -------
        list(numpy array (4x4))
            the path between the start and the goal Pose in SE(3) coordinates
        """

        #TODO: t3a-sampl - implement the PRM planner

        path = []

        print (start)
        print (goal)

        assert not environment.check_robot_collision(start), 'start collision'
        assert not environment.check_robot_collision(goal), 'goal collision'

        path = [start,goal]

        return(path)

